/* Gnome Calendar Go To Dialog Object
 * Copyright (C) 2002  Sebastian Rittau <srittau@jroger.in-berlin.de>
 *
 * $Id: gnomecal-goto.c,v 1.2 2002/08/24 17:02:01 srittau Exp $
 *
 * Based on the original go-to dialog by Federico Mena.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <time.h>

#include <glib.h>

#include <gconf/gconf.h>
#include <gconf/gconf-client.h>

#include <gnome.h>

#include "calendar-month.h"
#include "gnomecal-goto.h"
#include "gnomecal-marshal.h"
#include "utils-ui.h"


static void gncal_goto_init		(GncalGoto *dialog);
static void gncal_goto_finalize		(GObject *dialog);
static void gncal_goto_class_init	(GncalGotoClass *klass);
static void gncal_goto_setup		(GncalGoto *dialog);


struct _GncalGotoPriv {
	/* Widgets */

	GtkWidget *year_chooser;
	GtkWidget *month_chooser;
	GtkWidget *day_chooser;

	/* Signal IDs */

	gulong year_signal;
	gulong month_signal;
	gulong day_signal;
};

enum {
	SIGNAL_CHANGED,
	SIGNAL_YEAR_CHANGED,
	SIGNAL_MONTH_CHANGED,
	SIGNAL_DAY_CHANGED,
	SIGNAL_LAST
};


static GtkDialogClass *parent_class = NULL;

static guint signals[SIGNAL_LAST] = { 0 };


/*
 * Class Handling
 */

GType
gncal_goto_get_type (void)
{
	static GType gncal_goto_type = 0;

	if (!gncal_goto_type) {
		static const GTypeInfo gncal_goto_info = {
			sizeof (GncalGotoClass),
			NULL, /* base_init */
			NULL, /* base_finalize */
			(GClassInitFunc) gncal_goto_class_init,
			NULL, /* class_finalize */
			NULL, /* class_data */
			sizeof (GncalGoto),
			1,    /* n_preallocs */
			(GInstanceInitFunc) gncal_goto_init,
		};

		gncal_goto_type = g_type_register_static (GTK_TYPE_DIALOG,
							  "GncalGoto",
							  &gncal_goto_info,
							  0);
	}

	return gncal_goto_type;
}

static void
gncal_goto_class_init (GncalGotoClass *klass)
{
	GObjectClass *gobject_class;

	g_return_if_fail (klass != NULL);
	g_return_if_fail (GNCAL_IS_GOTO_CLASS (klass));

	gobject_class = G_OBJECT_CLASS (klass);

	gobject_class->finalize = gncal_goto_finalize;

	parent_class = g_type_class_peek_parent (klass);

	/* Signal setup */

	signals[SIGNAL_CHANGED] =
		g_signal_new ("changed",
			      G_TYPE_FROM_CLASS (klass),
			      G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
			      G_STRUCT_OFFSET (GncalGotoClass, changed),
			      NULL, NULL,
			      gnomecal_marshal_VOID__VOID,
			      G_TYPE_NONE, 0);
	signals[SIGNAL_YEAR_CHANGED] =
		g_signal_new ("year-changed",
			      G_TYPE_FROM_CLASS (klass),
			      G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
			      G_STRUCT_OFFSET (GncalGotoClass, year_changed),
			      NULL, NULL,
			      gnomecal_marshal_VOID__VOID,
			      G_TYPE_NONE, 0);
	signals[SIGNAL_MONTH_CHANGED] =
		g_signal_new ("month-changed",
			      G_TYPE_FROM_CLASS (klass),
			      G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
			      G_STRUCT_OFFSET (GncalGotoClass, month_changed),
			      NULL, NULL,
			      gnomecal_marshal_VOID__VOID,
			      G_TYPE_NONE, 0);
	signals[SIGNAL_DAY_CHANGED] =
		g_signal_new ("day-changed",
			      G_TYPE_FROM_CLASS (klass),
			      G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
			      G_STRUCT_OFFSET (GncalGotoClass, day_changed),
			      NULL, NULL,
			      gnomecal_marshal_VOID__VOID,
			      G_TYPE_NONE, 0);
}

static void
gncal_goto_init (GncalGoto *dialog)
{
	g_return_if_fail (dialog != NULL);
	g_return_if_fail (GNCAL_IS_GOTO (dialog));

	dialog->priv = g_new0 (GncalGotoPriv, 1);

	gncal_goto_setup (dialog);
}

static void
gncal_goto_finalize (GObject *object)
{
	GncalGoto *dialog;

	g_return_if_fail (object != NULL);
	g_return_if_fail (GNCAL_IS_GOTO (object));

	dialog = GNCAL_GOTO (object);

	g_free (dialog->priv);

	G_OBJECT_CLASS (parent_class)->finalize (object);
}

/*
 * Callbacks
 */

static void
gncal_goto_response (GtkDialog *dialog, gint response, gpointer data)
{
	g_return_if_fail (dialog != NULL);
	g_return_if_fail (GTK_IS_WINDOW (dialog));

	switch (response) {
	case GTK_RESPONSE_CLOSE:
		gtk_widget_destroy (GTK_WIDGET (dialog));
		break;
	case GTK_RESPONSE_DELETE_EVENT:
		break;
	case GTK_RESPONSE_HELP:
		ui_show_help (GTK_WINDOW (dialog), "goto-dialog");
		break;
	default:
		g_return_if_reached ();
	}
}

static void
gncal_goto_year_changed (GtkWidget *widget, GncalGoto *dialog)
{
	GncalGotoPriv *priv;

	struct tm tm;
	gint year;

	g_return_if_fail (dialog != NULL);
	g_return_if_fail (GNCAL_IS_GOTO (dialog));

	priv = dialog->priv;

	g_signal_handler_block (priv->day_chooser, priv->day_signal);

	calendar_month_get_date (CALENDAR_MONTH (priv->day_chooser), &tm);
	year = (gint) gtk_spin_button_get_value (GTK_SPIN_BUTTON (priv->year_chooser));
	tm.tm_year = year - 1900;
	calendar_month_set_date (CALENDAR_MONTH (priv->day_chooser), &tm);

	g_signal_handler_unblock (priv->day_chooser, priv->day_signal);

	g_signal_emit (dialog, signals[SIGNAL_YEAR_CHANGED], 0);
	g_signal_emit (dialog, signals[SIGNAL_CHANGED], 0);
}

static void
gncal_goto_month_changed (GtkWidget *widget, GncalGoto *dialog)
{
	GncalGotoPriv *priv;

	struct tm tm;
	gint month;

	g_return_if_fail (dialog != NULL);
	g_return_if_fail (GNCAL_IS_GOTO (dialog));

	priv = dialog->priv;

	g_signal_handler_block (priv->day_chooser, priv->day_signal);

	calendar_month_get_date (CALENDAR_MONTH (priv->day_chooser), &tm);
	month = gtk_option_menu_get_history (GTK_OPTION_MENU (priv->month_chooser));
	tm.tm_mon = month;
	calendar_month_set_date (CALENDAR_MONTH (priv->day_chooser), &tm);

	g_signal_handler_unblock (priv->day_chooser, priv->day_signal);

	g_signal_emit (dialog, signals[SIGNAL_MONTH_CHANGED], 0);
	g_signal_emit (dialog, signals[SIGNAL_CHANGED], 0);
}

static void
gncal_goto_day_changed (GtkWidget *widget, GncalGoto *dialog)
{
	g_return_if_fail (dialog != NULL);
	g_return_if_fail (GNCAL_IS_GOTO (dialog));

	g_signal_emit (dialog, signals[SIGNAL_DAY_CHANGED], 0);
	g_signal_emit (dialog, signals[SIGNAL_CHANGED], 0);
}

static void
gncal_goto_today_clicked (GtkWidget *widget, GncalGoto *dialog)
{
	g_return_if_fail (dialog != NULL);
	g_return_if_fail (GNCAL_IS_GOTO (dialog));

	gncal_goto_set_today (dialog);
}

/*
 * Setup
 */

static GtkWidget *
gncal_goto_month_om_new (GncalGoto *dialog)
{
	GtkWidget *om;
	GtkWidget *menu;

	gint i;

	static gchar *months[] = {
		N_("January"),
		N_("February"),
		N_("March"),
		N_("April"),
		N_("May"),
		N_("June"),
		N_("July"),
		N_("August"),
		N_("September"),
		N_("October"),
		N_("November"),
		N_("December")
	};

	om = gtk_option_menu_new ();

	menu = gtk_menu_new ();

	for (i = 0; i < G_N_ELEMENTS (months); i++) {
		GtkWidget *item;

		item = gtk_menu_item_new_with_label (_(months[i]));
		gtk_widget_show (item);
		gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
	}

	gtk_option_menu_set_menu (GTK_OPTION_MENU (om), menu);

	return om;
}

static void
gncal_goto_setup (GncalGoto *dialog)
{
	GncalGotoPriv	*priv;
	GtkWidget	*widget;
	GtkWidget	*box, *hbox;
	GtkWidget	*label;

	time_t t;
	struct tm tm;

	priv = dialog->priv;

	/* Widget Setup */

	gtk_window_set_title (GTK_WINDOW (dialog), _("Go to date"));

	gtk_dialog_add_buttons (GTK_DIALOG (dialog),
				GTK_STOCK_HELP,  GTK_RESPONSE_HELP,
				GTK_STOCK_CLOSE, GTK_RESPONSE_CLOSE,
				NULL);

	gtk_dialog_set_default_response (GTK_DIALOG (dialog), GTK_RESPONSE_CLOSE);

	g_signal_connect (G_OBJECT (dialog), "response",
			  G_CALLBACK (gncal_goto_response), NULL);

	box = GTK_DIALOG (dialog)->vbox;
	gtk_container_set_border_width (GTK_CONTAINER (box), GNOME_PAD_SMALL);
	gtk_box_set_spacing (GTK_BOX (box), GNOME_PAD_SMALL);

	/* Label */

	label = gtk_label_new (_("Please choose the date to jump to:"));
	gtk_widget_show (label);
	gtk_box_pack_start (GTK_BOX (box), label, FALSE, FALSE, 0);

	/* Year/Month */

	hbox = gtk_hbox_new (FALSE, GNOME_PAD_SMALL);
	gtk_widget_show (hbox);
	gtk_box_pack_start (GTK_BOX (box), hbox, FALSE, FALSE, 0);

	label = gtk_label_new_with_mnemonic (_("_Year:"));
	gtk_widget_show (label);
	gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);

	priv->year_chooser = gtk_spin_button_new_with_range (1900.0, 2199.0, 1.0);
	gtk_widget_set_size_request (priv->year_chooser, 60, -1);
	priv->year_signal =
		g_signal_connect (G_OBJECT (priv->year_chooser), "value-changed",
				  G_CALLBACK (gncal_goto_year_changed), dialog);
	gtk_widget_show (priv->year_chooser);
	gtk_box_pack_start (GTK_BOX (hbox), priv->year_chooser,
			    FALSE, FALSE, GNOME_PAD_SMALL);

	gtk_label_set_mnemonic_widget (GTK_LABEL (label), priv->year_chooser);

	label = gtk_label_new_with_mnemonic (_("_Month:"));
	gtk_widget_show (label);
	gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);

	priv->month_chooser = gncal_goto_month_om_new (dialog);
	priv->month_signal =
		g_signal_connect (G_OBJECT (priv->month_chooser), "changed",
				  G_CALLBACK (gncal_goto_month_changed), dialog);
	gtk_widget_show (priv->month_chooser);
	gtk_box_pack_start (GTK_BOX (hbox), priv->month_chooser, FALSE, FALSE, 0);

	gtk_label_set_mnemonic_widget (GTK_LABEL (label), priv->month_chooser);

	/* Calendar */

	priv->day_chooser = calendar_month_new ();
	priv->day_signal =
		g_signal_connect (G_OBJECT (priv->day_chooser),
				  "selection-changed",
				  G_CALLBACK (gncal_goto_day_changed), dialog);
	gtk_widget_show (priv->day_chooser);
	gtk_box_pack_start (GTK_BOX (box), priv->day_chooser, TRUE, TRUE, 0);

	/* Today button */

	hbox = gtk_hbox_new (FALSE, 0);
	gtk_widget_show (hbox);
	gtk_box_pack_start (GTK_BOX (box), hbox, FALSE, FALSE, 0);

	widget = gtk_button_new_with_mnemonic (_("_Today"));
	g_signal_connect (G_OBJECT (widget), "clicked",
			  G_CALLBACK (gncal_goto_today_clicked), dialog);
	gtk_widget_show (widget);
	gtk_box_pack_start (GTK_BOX (hbox), widget, FALSE, FALSE, 0);

	/* Setup defaults */

	t = time (NULL);
	localtime_r (&t, &tm);

	gtk_spin_button_set_value (GTK_SPIN_BUTTON (priv->year_chooser), tm.tm_year + 1900.0);
	gtk_option_menu_set_history (GTK_OPTION_MENU (priv->month_chooser), tm.tm_mon);
	calendar_month_set_date (CALENDAR_MONTH (priv->day_chooser), &tm);
}

/*
 * Exported Methods
 */

GtkWidget *
gncal_goto_new (GtkWindow *parent)
{
	GncalGoto *dialog;

	g_return_val_if_fail (parent == NULL || GTK_IS_WINDOW (parent), NULL);

	dialog = g_object_new (GNCAL_TYPE_GOTO, NULL);

	if (parent)
		gtk_window_set_transient_for (GTK_WINDOW (dialog), parent);

	return GTK_WIDGET (dialog);
}

void gncal_goto_set_date (GncalGoto *dialog, const struct tm *tm)
{
	GncalGotoPriv *priv;

	guint changed;
	struct tm mytm;

	g_return_if_fail (dialog != NULL);
	g_return_if_fail (GNCAL_IS_GOTO (dialog));

	priv = dialog->priv;

	g_signal_handler_block (priv->year_chooser,  priv->year_signal);
	g_signal_handler_block (priv->month_chooser, priv->month_signal);
	g_signal_handler_block (priv->day_chooser,   priv->day_signal);

	calendar_month_get_date (CALENDAR_MONTH (priv->day_chooser), &mytm);

	changed = 0;

	if (mytm.tm_year != tm->tm_year) {
		gtk_spin_button_set_value (GTK_SPIN_BUTTON (priv->year_chooser), tm->tm_year + 1900.0);
		changed |= 0x1;
	}
	if (mytm.tm_mon != tm->tm_mon) {
		gtk_option_menu_set_history (GTK_OPTION_MENU (priv->month_chooser), tm->tm_mon);
		changed |= 0x2;
	}
	if (mytm.tm_mday != tm->tm_mday)
		changed |= 0x4;

	if (changed)
		calendar_month_set_date (CALENDAR_MONTH (priv->day_chooser), tm);

	g_signal_handler_unblock (priv->year_chooser,  priv->year_signal);
	g_signal_handler_unblock (priv->month_chooser, priv->month_signal);
	g_signal_handler_unblock (priv->day_chooser,   priv->day_signal);

	if (changed & 0x1)
		g_signal_emit (dialog, signals[SIGNAL_YEAR_CHANGED], 0);
	if (changed & 0x2)
		g_signal_emit (dialog, signals[SIGNAL_MONTH_CHANGED], 0);
	if (changed & 0x4)
		g_signal_emit (dialog, signals[SIGNAL_DAY_CHANGED], 0);
	if (changed)
		g_signal_emit (dialog, signals[SIGNAL_CHANGED], 0);
}

void gncal_goto_set_today (GncalGoto *dialog)
{
	time_t t;
	struct tm tm;

	g_return_if_fail (dialog != NULL);
	g_return_if_fail (GNCAL_IS_GOTO (dialog));

	t = time (NULL);
	localtime_r (&t, &tm);

	gncal_goto_set_date (dialog, &tm);
}

void gncal_goto_get_date (GncalGoto *dialog, struct tm *tm)
{
	GncalGotoPriv *priv;

	g_return_if_fail (dialog != NULL);
	g_return_if_fail (GNCAL_IS_GOTO (dialog));
	g_return_if_fail (tm != NULL);

	priv = dialog->priv;

	calendar_month_get_date (CALENDAR_MONTH (priv->day_chooser), tm);
}
