/* Quick view widget for Gnomecal
 *
 * Copyright (C) 1998 The Free Software Foundation
 *
 * Author: Federico Mena <federico@nuclecu.unam.mx>
 *
 * Changed in 2002 by Sebastian Rittau <srittau@jroger.in-berlin.de>
 * for the GNOME 2 port.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <mimedir/mimedir-vevent.h>

#include "quick-view.h"
#include "utils-time.h"


#define QUICK_VIEW_FONT "Sans 10"


static void quick_view_class_init (QuickViewClass *class, gpointer data);
static void quick_view_init       (QuickView      *qv);
static void quick_view_destroy    (GtkObject      *object);

static gint quick_view_button_release (GtkWidget *widget, GdkEventButton *event);
static gint quick_view_map_event (GtkWidget *widget, GdkEventAny *event);


static GtkWindowClass *parent_class;


GType
quick_view_get_type (void)
{
	static GType quick_view_type = 0;

	if (!quick_view_type) {
		static const GTypeInfo quick_view_info = {
			sizeof (QuickViewClass),
			NULL, /* base_init */
			NULL, /* base_finalize */
			(GClassInitFunc) quick_view_class_init,
			NULL, /* class_finalize */
			NULL, /* class_data */
			sizeof (QuickView),
			4,    /* n_preallocs */
			(GInstanceInitFunc) quick_view_init,
		};

		quick_view_type = g_type_register_static (GTK_TYPE_WINDOW,
							  "QuickView",
							  &quick_view_info,
							  0);
	}

	return quick_view_type;
}

static void
quick_view_class_init (QuickViewClass *class, gpointer data)
{
	GtkObjectClass *gtkobject_class = GTK_OBJECT_CLASS (class);
	GtkWidgetClass *widget_class    = GTK_WIDGET_CLASS (class);

	parent_class = g_type_class_peek_parent (class);

	gtkobject_class->destroy = quick_view_destroy;

	widget_class->button_release_event = quick_view_button_release;
	widget_class->map_event = quick_view_map_event;
}

static void
quick_view_init (QuickView *qv)
{
	GtkWindow *win;

	g_return_if_fail (qv != NULL);
	g_return_if_fail (IS_QUICK_VIEW (qv));

	win = GTK_WINDOW (qv);

	win->type = GTK_WINDOW_POPUP;
	gtk_window_set_position  (win, GTK_WIN_POS_MOUSE);
	gtk_window_set_decorated (win, FALSE);

	qv->calendar = NULL;
	qv->main_loop = g_main_loop_new (NULL, FALSE);
}

static void
quick_view_destroy (GtkObject *object)
{
	QuickView *qv;

	g_return_if_fail (object != NULL);
	g_return_if_fail (IS_QUICK_VIEW (object));

	qv = QUICK_VIEW (object);

	if (qv->main_loop) /* FIXME: needed? */
		g_main_loop_unref (qv->main_loop);
	qv->main_loop = NULL;

	GTK_OBJECT_CLASS (parent_class)->destroy (object);
}

static gint
quick_view_button_release (GtkWidget *widget, GdkEventButton *event)
{
	QuickView *qv;

	g_return_val_if_fail (widget != NULL, TRUE);
	g_return_val_if_fail (IS_QUICK_VIEW (widget), TRUE);

	qv = QUICK_VIEW (widget);

	if (event->button != qv->button)
		return FALSE;

	gdk_pointer_ungrab (event->time);
	gtk_grab_remove (GTK_WIDGET (qv));
	gtk_widget_hide (GTK_WIDGET (qv));

	g_main_loop_quit (qv->main_loop); /* End modality of the quick view */

	return TRUE;
}

static gint
quick_view_map_event (GtkWidget *widget, GdkEventAny *event)
{
	GdkCursor *cursor;

	g_return_val_if_fail (widget != NULL, FALSE);
	g_return_val_if_fail (IS_QUICK_VIEW (widget), FALSE);

	cursor = gdk_cursor_new (GDK_ARROW);
	gdk_pointer_grab (widget->window,
			  TRUE,
			  GDK_BUTTON_RELEASE_MASK,
			  NULL,
			  cursor,
			  GDK_CURRENT_TIME);
	gdk_cursor_unref (cursor);

	return FALSE;
}

/* Creates the items corresponding to a single calendar object.  Takes in the y position of the
 * items to create and returns the y position of the next item to create.  Also takes in the current
 * maximum width for items and returns the new maximum width.
 * This is used in the popup-window in the year view
 */
static void
create_items_for_event (QuickView *qv, CalendarObject *co, double *y, double *max_width)
{
	GnomeCanvas *canvas;
	GnomeCanvasItem *item;
	struct tm start_tm, end_tm;
	gchar *str;
	double text_width, text_height;

	g_return_if_fail (qv != NULL);
	g_return_if_fail (IS_QUICK_VIEW (qv));

	/* FIXME: make this nice */

	canvas = GNOME_CANVAS (qv->canvas);

	start_tm = *localtime (&co->ev_start);
	end_tm   = *localtime (&co->ev_end);

	if (mimedir_vcomponent_get_allday (MIMEDIR_VCOMPONENT (co->event))) {
		g_object_get (G_OBJECT (co->event), "summary", str, NULL);
	} else {
		gchar *start, *end;
		gchar *summary;

		start = time_get_short_time_string (&start_tm);
		end   = time_get_short_time_string (&end_tm);

		g_object_get (G_OBJECT (co->event), "summary", &summary, NULL);

		/* Translators: STARTDATE - ENDDATE SUMMARY */
		str = g_strdup_printf (_("%s - %s %s"), start, end, summary);

		g_free (summary);
		g_free (start);
		g_free (end);
	}

	item = gnome_canvas_item_new (gnome_canvas_root (canvas),
				      gnome_canvas_text_get_type (),
				      "x", 0.0,
				      "y", *y,
				      "anchor", GTK_ANCHOR_NW,
				      "text", str,
				      "font", QUICK_VIEW_FONT,
				      NULL);

	g_free (str);

	/* Measure the text and return the proper size values */

	g_object_get (G_OBJECT (item),
		      "text_width",  &text_width,
		      "text_height", &text_height,
		      NULL);

	if (text_width > *max_width)
		*max_width = text_width;

	*y += text_height;
}

/* Creates the canvas items corresponding to the events in the list */
static void
setup_event_list (QuickView *qv, GList *event_list)
{
	CalendarObject *co;
	GnomeCanvasItem *item;
	double y, max_width;

	g_return_if_fail (qv != NULL);
	g_return_if_fail (IS_QUICK_VIEW (qv));

	/* If there are no events, then just put a simple label */

	if (!event_list) {
		item = gnome_canvas_item_new (gnome_canvas_root (GNOME_CANVAS (qv->canvas)),
					      gnome_canvas_text_get_type (),
					      "x", 0.0,
					      "y", 0.0,
					      "anchor", GTK_ANCHOR_NW,
					      "text", _("No appointments for this day"),
					      "font", QUICK_VIEW_FONT,
					      NULL);

		/* Measure the text and set the proper sizes */

		g_object_get (G_OBJECT (item),
			      "text_width",  &max_width,
			      "text_height", &y,
			      NULL);
	} else {
		/* Create the items for all the events in the list */

		y = 0.0;
		max_width = 0.0;

		for (; event_list; event_list = event_list->next) {
			co = event_list->data;
			create_items_for_event (qv, co, &y, &max_width);
		}
	}

	/* Set the scrolling region to fit all the items */

	gnome_canvas_set_scroll_region (GNOME_CANVAS (qv->canvas),
					0.0, 0.0,
					max_width, y);

	gtk_widget_set_size_request (qv->canvas, max_width, y);
}

GtkWidget *
quick_view_new (GnomeCalendar *calendar, char *title, GList *event_list)
{
	QuickView *qv;
	GtkWidget *vbox;
	GtkWidget *w;

	g_return_val_if_fail (calendar != NULL, NULL);
	g_return_val_if_fail (GNOME_IS_CALENDAR (calendar), NULL);

	qv = g_object_new (TYPE_QUICK_VIEW, NULL);
	qv->calendar = calendar;

	/* Create base widgets for the popup window */

	w = gtk_frame_new (NULL);
	gtk_frame_set_shadow_type (GTK_FRAME (w), GTK_SHADOW_ETCHED_OUT);
	gtk_widget_show (w);
	gtk_container_add (GTK_CONTAINER (qv), w);

	vbox = gtk_vbox_new (FALSE, 0);
	gtk_widget_show (vbox);
	gtk_container_add (GTK_CONTAINER (w), vbox);

	w = gtk_label_new (title);
	gtk_widget_show (w);
	gtk_box_pack_start (GTK_BOX (vbox), w, FALSE, FALSE, 0);

	w = gtk_hseparator_new ();
	gtk_widget_show (w);
	gtk_box_pack_start (GTK_BOX (vbox), w, FALSE, FALSE, 0);

	w = gtk_alignment_new (0.5, 0.5, 0.0, 0.0);
	gtk_container_set_border_width (GTK_CONTAINER (w), GNOME_PAD_SMALL);
	gtk_widget_show (w);
	gtk_box_pack_start (GTK_BOX (vbox), w, TRUE, TRUE, 0);

	gtk_widget_push_colormap (gdk_colormap_get_system ());

	qv->canvas = gnome_canvas_new ();
	gtk_widget_show (qv->canvas);

	gtk_widget_pop_colormap ();

	gtk_container_add (GTK_CONTAINER (w), qv->canvas);

	/* Set up the event list */

	setup_event_list (qv, event_list);

	return GTK_WIDGET (qv);
}

void
quick_view_do_popup (QuickView *qv, GdkEventButton *event)
{
	g_return_if_fail (qv != NULL);
	g_return_if_fail (IS_QUICK_VIEW (qv));
	g_return_if_fail (event != NULL);

	/* Pop up the window */

	gtk_widget_show (GTK_WIDGET (qv));
	gtk_grab_add (GTK_WIDGET (qv));

	qv->button = event->button;

	g_main_loop_run (qv->main_loop); /* Begin modality */

	/* The button release event handler will call g_main_loop_quit() */
}
