/* Year view display for gncal
 *
 * Copyright (C) 1998 The Free Software Foundation
 *
 * Authors: Arturo Espinosa <arturo@nuclecu.unam.mx>
 *          Federico Mena <federico@nuclecu.unam.mx>
 *
 * Changed in 2002 and 2003 by Sebastian Rittau <srittau@jroger.in-berlin.de>
 * for the GNOME 2 port.
 *
 * $Id: year-view.c,v 1.42 2003/01/11 02:48:08 srittau Exp $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string.h>
#include <time.h>

#include <gconf/gconf.h>
#include <gconf/gconf-client.h>
#include <gnome.h>

#include "eventedit.h"
#include "gnomecal-config.h"
#include "mark.h"
#include "quick-view.h"
#include "timeutil.h"
#include "utils-time.h"
#include "year-view.h"


#define HEAD_SPACING 4		/* Spacing between year heading and months */
#define TITLE_SPACING 1		/* Spacing between title and calendar */
#define SPACING 4		/* Spacing between months */


static void year_view_class_init	(YearViewClass	*klass);
static void year_view_init		(YearView	*yv);
static void year_view_destroy		(GtkObject	*object);
static void year_view_finalize		(GObject	*object);
static void year_view_size_request	(GtkWidget	*widget,
					 GtkRequisition	*requisition);
static void year_view_size_allocate	(GtkWidget	*widget,
					 GtkAllocation	*allocation);


struct _YearViewPriv {
	gchar *color_fg, *color_current;

	/* Signal IDs */

	guint notify1, notify2;
};


static GnomeCanvas *parent_class;


GType
year_view_get_type (void)
{
	static GType year_view_type = 0;

	if (!year_view_type) {
		static const GTypeInfo year_view_info = {
			sizeof (YearViewClass),
			NULL, /* base_init */
			NULL, /* base_finalize */
			(GClassInitFunc) year_view_class_init,
			NULL, /* class_finalize */
			NULL, /* class_data */
			sizeof (YearView),
			1,    /* n_preallocs */
			(GInstanceInitFunc) year_view_init,
		};

		year_view_type = g_type_register_static (GNOME_TYPE_CANVAS,
							 "YearView",
							 &year_view_info,
							 0);
	}

	return year_view_type;
}

static void
year_view_class_init (YearViewClass *klass)
{
	GObjectClass   *gobject_class;
	GtkObjectClass *object_class;
	GtkWidgetClass *widget_class;

	gobject_class = G_OBJECT_CLASS   (klass);
	object_class  = GTK_OBJECT_CLASS (klass);
	widget_class  = GTK_WIDGET_CLASS (klass);

	parent_class = g_type_class_peek_parent (klass);

	gobject_class->finalize = year_view_finalize;

	object_class->destroy = year_view_destroy;

	widget_class->size_request = year_view_size_request;
	widget_class->size_allocate = year_view_size_allocate;
}

static void
year_view_finalize (GObject *object)
{
	YearView *yv;

	g_return_if_fail (object != NULL);
	g_return_if_fail (IS_YEAR_VIEW (object));

	yv = YEAR_VIEW (object);

	g_free (yv->priv->color_fg);
	g_free (yv->priv->color_current);

	g_free (yv->priv);

	G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
year_view_destroy (GtkObject *object)
{
	YearView *yv;
	GConfClient *gconf;

	g_return_if_fail (object != NULL);
	g_return_if_fail (IS_YEAR_VIEW (object));

	yv = YEAR_VIEW (object);

	gconf = gconf_client_get_default ();

	if (yv->priv->notify1) {
		gconf_client_notify_remove (gconf, yv->priv->notify1);
		yv->priv->notify1 = 0;
	}
	if (yv->priv->notify2) {
		gconf_client_notify_remove (gconf, yv->priv->notify2);
		yv->priv->notify2 = 0;
	}

	g_object_unref (G_OBJECT (gconf));

	if (yv->need_resize) {
		yv->need_resize = FALSE;
		gtk_idle_remove (yv->idle_id);
	}

	GTK_OBJECT_CLASS (parent_class)->destroy (object);
}

/*
 * Callbacks
 */

static void mark_current_day (YearView *yv);

static void
year_view_config_changed (GConfClient *gconf,
			  guint cnxn_id,
			  GConfEntry *entry,
			  gpointer data)
{
	YearView *yv;
	const gchar *key, *value;

	g_return_if_fail (data != NULL);
	g_return_if_fail (IS_YEAR_VIEW (data));

	yv = YEAR_VIEW (data);

	key   = gconf_entry_get_key (entry);
	value = gconf_value_get_string (gconf_entry_get_value (entry));

	if (!strcmp (key, CONFIG_COLOR_DAY_NUMBERS)) {
		g_free (yv->priv->color_fg);
		yv->priv->color_fg = g_strdup (value);
	} else if (!strcmp (key, CONFIG_COLOR_CURRENT_DAY)) {
		g_free (yv->priv->color_current);
		yv->priv->color_current = g_strdup (value);
	} else
		g_return_if_reached ();

	mark_current_day (yv);
}

/* Resizes the year view's child items.  This is done in the idle loop for
 * performance (we avoid resizing on every size allocation).
 */
static gint
idle_handler (gpointer data)
{
	YearView *yv;
	double head_height;
	double title_height;
	double width, height;
	double month_width;
	double month_height;
	double month_yofs;
	double xofs, yofs;
	double x, y;
	int i;

	yv = data;

	/* Compute the size we can use */

	width = MAX (GTK_WIDGET (yv)->allocation.width, yv->min_width);
	height = MAX (GTK_WIDGET (yv)->allocation.height, yv->min_height);

	gnome_canvas_set_scroll_region (GNOME_CANVAS (yv), 0, 0, width, height);

	width--;
	height--;

	/* Get the heights of the heading and the titles */

	g_object_get (GTK_OBJECT (yv->heading),
		      "text_height", &head_height,
		      NULL);
	head_height += 2 * HEAD_SPACING;

	g_object_get (GTK_OBJECT (yv->titles[0]),
		      "text_height", &title_height,
		      NULL);

	/* Offsets */

	xofs = (width + SPACING) / 3.0;
	yofs = (height - head_height + SPACING) / 4.0;

	/* Month item vertical offset */

	month_yofs = title_height + TITLE_SPACING;

	/* Month item dimensions */

	month_width = (width - 2 * SPACING) / 3.0;
	month_height = (yofs - SPACING) - month_yofs;

	/* Adjust the year heading */

	gnome_canvas_item_set (yv->heading,
			       "x", width / 2.0,
			       "y", (double) HEAD_SPACING,
			       NULL);

	/* Adjust titles and months */

	for (i = 0; i < 12; i++) {
		x = (i % 3) * xofs;
		y = head_height + 12 + (i / 3) * yofs;

		gnome_canvas_item_set (yv->titles[i],
				       "x", x + month_width / 2.0,
				       "y", y,
				       NULL);

		gnome_canvas_item_set (yv->mitems[i],
				       "x", x,
				       "y", y + month_yofs,
				       "width", month_width,
				       "height", month_height,
				       NULL);
	}

	/* Done */

	yv->need_resize = FALSE;
	return FALSE;
}

/* Marks the year view as needing a resize, which will be performed during the idle loop */
static void
need_resize (YearView *yv)
{
	if (yv->need_resize)
		return;

	yv->need_resize = TRUE;
	yv->idle_id = gtk_idle_add (idle_handler, yv);
}

/* Callback used to destroy the year view's popup menu when the year view itself is destroyed */
static void
destroy_menu (GtkWidget *widget, gpointer data)
{
	gtk_widget_destroy (GTK_WIDGET (data));
}

/* Create a new appointment in the highlighted day from the year view's popup menu */
static void
new_appointment (GtkWidget *widget, gpointer data)
{
	GtkWidget *ce;
	YearView *yv;
	time_t *t;

	yv = YEAR_VIEW (data);
	t = g_object_get_data (G_OBJECT (widget), "time_data");

	ce = event_editor_new_whole_day (*t);

	gtk_widget_show (ce);
}

/* Convenience functions to jump to a view and set the time */
static void
do_jump (GtkWidget *widget, gpointer data, char *view_name)
{
	YearView *yv;
	time_t *t;

	yv = YEAR_VIEW (data);

	/* Get the time data from the menu item */

	t = g_object_get_data (G_OBJECT (widget), "time_data");

	/* Set the view and time */

	gnome_calendar_set_view (yv->calendar, view_name);
	gnome_calendar_goto (yv->calendar, *t);
}

/* The following three callbacks set the view in the calendar and change the time */

static void
jump_to_day (GtkWidget *widget, gpointer data)
{
	do_jump (widget, data, "dayview");
}

static void
jump_to_week (GtkWidget *widget, gpointer data)
{
	do_jump (widget, data, "weekview");
}

static void
jump_to_month (GtkWidget *widget, gpointer data)
{
	do_jump (widget, data, "monthview");
}

/* Information for the year view's popup menu */
static GnomeUIInfo yv_popup_menu[] = {
	GNOMEUIINFO_ITEM_STOCK (N_("_New appointment in this day..."), NULL, new_appointment, GTK_STOCK_NEW),

	GNOMEUIINFO_SEPARATOR,

	GNOMEUIINFO_ITEM_STOCK (N_("Jump to this _day"), NULL, jump_to_day, GTK_STOCK_JUMP_TO),
	GNOMEUIINFO_ITEM_STOCK (N_("Jump to this _week"), NULL, jump_to_week, GTK_STOCK_JUMP_TO),
	GNOMEUIINFO_ITEM_STOCK (N_("Jump to this _month"), NULL, jump_to_month, GTK_STOCK_JUMP_TO),
	GNOMEUIINFO_END
};

/* Returns the popup menu cooresponding to the specified year view.  If the menu has not been
 * created yet, it creates it and attaches it to the year view.
 */
static GtkWidget *
get_popup_menu (YearView *yv)
{
	GtkWidget *menu;

	menu = g_object_get_data (G_OBJECT (yv), "popup_menu");

	if (!menu) {
		menu = gnome_popup_menu_new (yv_popup_menu);
		g_object_set_data (G_OBJECT (yv), "popup_menu", menu);
		g_signal_connect (G_OBJECT (yv), "destroy",
				  G_CALLBACK (destroy_menu),
				  menu);
	}

	return menu;
}

/* Executes the year view's popup menu.  It may disable/enable some menu items based on the
 * specified flags.  A pointer to a time_t value containing the specified time data is set in the
 * "time_data" object data key of the menu items.
 */
static void
do_popup_menu (YearView *yv, GdkEventButton *event, int allow_new, int allow_day, int allow_week, int allow_month,
	       int year, int month, int day)
{
	GtkWidget *menu;
	static time_t t;

	menu = get_popup_menu (yv);

	/* Enable/disable items as appropriate */

	gtk_widget_set_sensitive (yv_popup_menu[0].widget, allow_new);
	gtk_widget_set_sensitive (yv_popup_menu[2].widget, allow_day);
	gtk_widget_set_sensitive (yv_popup_menu[3].widget, allow_week);
	gtk_widget_set_sensitive (yv_popup_menu[4].widget, allow_month);

	/* Set the day item relevant to the context */

	t = time_from_day (year, month, day);

	g_object_set_data (G_OBJECT (yv_popup_menu[0].widget), "time_data", &t);
	g_object_set_data (G_OBJECT (yv_popup_menu[2].widget), "time_data", &t);
	g_object_set_data (G_OBJECT (yv_popup_menu[3].widget), "time_data", &t);
	g_object_set_data (G_OBJECT (yv_popup_menu[4].widget), "time_data", &t);

	gnome_popup_menu_do_popup (menu, NULL, NULL, event, yv, GTK_WIDGET (yv));
}

/* Creates the quick view when the user clicks on a day */
static void
do_quick_view_popup (YearView *yv, GdkEventButton *event, int year, int month, int day)
{
	time_t day_start, day_end;
	GList *list;
	GtkWidget *qv;
	gchar *s;

	day_start = time_day_begin (time_from_day (year, month, day));
	day_end = time_day_end (day_start);

	list = calendar_get_events_in_range (yv->calendar->cal, day_start, day_end);

	s = time_get_date_string (localtime (&day_start));
	qv = quick_view_new (yv->calendar, s, list);
	g_free (s);

	quick_view_do_popup (QUICK_VIEW (qv), event);

	gtk_widget_destroy (qv);
	calendar_destroy_event_list (list);
}

/* Event handler for days in the year's month items */
static gint
day_event (GnomeCanvasItem *item, GdkEvent *event, gpointer data)
{
	YearView *yv;
	GnomeMonthItem *mitem;
	int child_num;
	guint year, month, day;

	mitem = GNOME_MONTH_ITEM (data);
	child_num = gnome_month_item_child2num (mitem, item);
	day = gnome_month_item_num2day (mitem, child_num);

	g_object_get (G_OBJECT (mitem),
		      "year",  &year,
		      "month", &month,
		      NULL);

	yv = YEAR_VIEW (item->canvas);

	switch (event->type) {
	case GDK_BUTTON_PRESS:
		if (day == 0)
			break;

		if (event->button.button == 1) {
			do_quick_view_popup (yv, (GdkEventButton *) event,
					     year, month, day);
			return TRUE;
		} else if (event->button.button == 3) {
			do_popup_menu (yv, (GdkEventButton *) event,
				       TRUE, TRUE, TRUE, TRUE,
				       year, month, day);

			/* We have to stop the signal emission because mark.c will grab it too and
			 * set the return value to FALSE.  Blargh.
			 */
			g_signal_stop_emission_by_name (G_OBJECT (item),
							"event");
			return TRUE;
		}

		break;

	default:
		break;
	}

	return FALSE;
}

/* Event handler for whole month items */
static gint
month_event (GnomeCanvasItem *item, GdkEvent *event, gpointer data)
{
	YearView *yv;
	GnomeMonthItem *mitem;
	guint year, month;

	mitem = GNOME_MONTH_ITEM (item);
	yv    = YEAR_VIEW (item->canvas);

	g_object_get (G_OBJECT (mitem),
		      "year",  &year,
		      "month", &month,
		      NULL);

	switch (event->type) {
	case GDK_BUTTON_PRESS:
		if (event->button.button != 3)
			break;

		do_popup_menu (yv, (GdkEventButton *) event,
			       FALSE, FALSE, FALSE, TRUE,
			       year, month, 1);

		/* We have to stop the signal emission because mark.c will grab it too and
		 * set the return value to FALSE.  Blargh.
		 */
		g_signal_stop_emission_by_name (G_OBJECT (item), "event");
		return TRUE;

	default:
		break;
	}

	return FALSE;
}

/* Sets up the month item with the specified index -- connects signals for handling events, etc. */
static void
setup_month_item (YearView *yv, int n)
{
	GnomeCanvasItem *mitem;
	GnomeCanvasItem *item;
	int i;

	mitem = yv->mitems[n];

	/* Connect the day signals */

	for (i = 0; i < 42; i++) {
		item = gnome_month_item_num2child (GNOME_MONTH_ITEM (mitem), GNOME_MONTH_ITEM_DAY_GROUP + i);
		g_signal_connect (G_OBJECT (item), "event",
				  G_CALLBACK (day_event), mitem);
	}

	/* Connect the month signals */

	g_signal_connect (G_OBJECT (mitem), "event",
			  G_CALLBACK (month_event), NULL);
}

/* Computes the minimum size for the year view and stores it in its internal fields */
static void
compute_min_size (YearView *yv)
{
	double m_width;
	double m_height;
	double max_width;
	double w;
	double d, width, height;
	int i;

	/* Compute the minimum size of the year heading */

	g_object_get (GTK_OBJECT (yv->heading),
		      "text_width",  &m_width,
		      "text_height", &m_height,
		      NULL);

	m_height += 2 * HEAD_SPACING;

	/* Add spacing */

	m_height += 12;

	/* Add height of month titles and their spacings */

	g_object_get (GTK_OBJECT (yv->titles[0]),
		      "text_height", &d,
		      NULL);

	m_height += 4 * (d + TITLE_SPACING);

	/* Add width of month titles */

	max_width = 0.0;

	for (i = 0; i < 12; i++) {
		g_object_get (GTK_OBJECT (yv->titles[i]),
			      "text_width", &w,
			      NULL);

		if (max_width < w)
			max_width = w;
	}

	max_width = 3 * max_width + 2 * SPACING;

	if (m_width < max_width)
		m_width = max_width;

	/* Add width of month items */

	g_object_get (GTK_OBJECT (yv->mitems[0]),
		      "width",  &width,
		      "height", &height,
		      NULL);

	max_width = 3 * width + 2 * SPACING;

	if (m_width < max_width)
		m_width = max_width;

	/* Add height of month items */

	m_height += 4 * height + 3 * SPACING;

	/* Finally, set the minimum width and height in the year view */

	yv->min_width = (int) (m_width + 0.5);
	yv->min_height = (int) (m_height + 0.5);
}

static void
year_view_reordered (GnomeMonthItem *item, YearView *yv)
{
	g_return_if_fail (item != NULL);
	g_return_if_fail (GNOME_IS_MONTH_ITEM (item));
	g_return_if_fail (yv != NULL);
	g_return_if_fail (IS_YEAR_VIEW (yv));

	unmark_month_item (item);
	mark_month_item (item, yv->calendar->cal);
	mark_current_day (yv);
}

static void
year_view_init (YearView *yv)
{
	YearViewPriv *priv;
	GConfClient *gconf;
	int i;
	char buf[100];
	struct tm tm;

	priv = g_new0 (YearViewPriv, 1);
	yv->priv = priv;

	memset (&tm, 0, sizeof (tm));

	/* Heading */

	yv->heading = gnome_canvas_item_new (gnome_canvas_root (GNOME_CANVAS (yv)),
					     gnome_canvas_text_get_type (),
					     "anchor", GTK_ANCHOR_N,
					     "font", HEADING_FONT,
					     "fill_color", "black",
					     NULL);

	/* Months */

	for (i = 0; i < 12; i++) {
		gchar *s;

		/* Title */

		strftime (buf, sizeof (buf), "%B", &tm);
		s = g_locale_to_utf8 (buf, -1, NULL, NULL, NULL);
		
		tm.tm_mon++;

		yv->titles[i] = gnome_canvas_item_new (gnome_canvas_root (GNOME_CANVAS (yv)),
						       gnome_canvas_text_get_type (),
						       "text", s,
						       "anchor", GTK_ANCHOR_N,
						       "font", TITLE_FONT,
						       "fill_color", "black",
						       NULL);
		g_free (s);

		/* Month item */

		yv->mitems[i] = gnome_month_item_new (gnome_canvas_root (GNOME_CANVAS (yv)));
		gnome_canvas_item_set (yv->mitems[i],
				       "anchor", GTK_ANCHOR_NW,
				       "heading_font", DAY_HEADING_FONT,
				       "day_font", NORMAL_DAY_FONT,
				       NULL);
		g_signal_connect (G_OBJECT (yv->mitems[i]), "reordered",
				  G_CALLBACK (year_view_reordered), yv);
		setup_month_item (yv, i);
	}

	/* We will need to resize the items when we paint for the first time */

	yv->old_marked_day = -1;
	yv->idle_id = -1;
	need_resize (yv);

	/* Configuration */

	gconf = gconf_client_get_default ();

	gconf_client_add_dir (gconf, CONFIG_COLOR_DIR, GCONF_CLIENT_PRELOAD_NONE, NULL);

	priv->color_fg      = gncal_config_get_color (gconf, CONFIG_COLOR_DAY_NUMBERS, DEFAULT_COLOR_DAY_NUMBERS);
	priv->color_current = gncal_config_get_color (gconf, CONFIG_COLOR_CURRENT_DAY, DEFAULT_COLOR_CURRENT_DAY);

	priv->notify1 = gconf_client_notify_add (gconf,
						 CONFIG_COLOR_DAY_NUMBERS,
						 year_view_config_changed,
						 yv, NULL, NULL);
	priv->notify2 = gconf_client_notify_add (gconf,
						 CONFIG_COLOR_CURRENT_DAY,
						 year_view_config_changed,
						 yv, NULL, NULL);

	g_object_unref (G_OBJECT (gconf));
}

GtkWidget *
year_view_new (GnomeCalendar *calendar, time_t year)
{
	YearView *yv;

	g_return_val_if_fail (calendar != NULL, NULL);
	g_return_val_if_fail (GNOME_IS_CALENDAR (calendar), NULL);

	yv = g_object_new (TYPE_YEAR_VIEW, NULL);
	yv->calendar = calendar;

	year_view_set (yv, year);
	compute_min_size (yv);

	return GTK_WIDGET (yv);
}

static void
year_view_size_request (GtkWidget *widget, GtkRequisition *requisition)
{
	YearView *yv;

	g_return_if_fail (widget != NULL);
	g_return_if_fail (IS_YEAR_VIEW (widget));
	g_return_if_fail (requisition != NULL);

	yv = YEAR_VIEW (widget);

	requisition->width = yv->min_width;
	requisition->height = yv->min_height;
}

static void
year_view_size_allocate (GtkWidget *widget, GtkAllocation *allocation)
{
	YearView *yv;

	g_return_if_fail (widget != NULL);
	g_return_if_fail (IS_YEAR_VIEW (widget));
	g_return_if_fail (allocation != NULL);

	yv = YEAR_VIEW (widget);

	if (GTK_WIDGET_CLASS (parent_class)->size_allocate)
		(* GTK_WIDGET_CLASS (parent_class)->size_allocate) (widget, allocation);

	need_resize (yv);
}

void
year_view_update (YearView *yv, MIMEDirVComponent *comp, int flags)
{
	g_return_if_fail (yv != NULL);
	g_return_if_fail (IS_YEAR_VIEW (yv));

	/* If only the summary changed, we don't care */

	if (comp && ((flags & CHANGE_SUMMARY) == flags))
		return;

	year_view_set (yv, time_year_begin (time_from_day (yv->year, 0, 1)));
}

/* Unmarks the old day that was marked as current and marks the current day if appropriate */
static void
mark_current_day (YearView *yv)
{
	time_t t;
	struct tm tm;
	int month_index, day_index;
	GnomeCanvasItem *item;

	/* Unmark the old day */

	if (yv->old_marked_day != -1) {
		const gchar *color;

		color = yv->priv->color_fg;
		if (!pango_color_parse (NULL, color))
			color = "#000";

		month_index = yv->old_marked_day / 42;
		day_index = yv->old_marked_day % 42;

		item = gnome_month_item_num2child (GNOME_MONTH_ITEM (yv->mitems[month_index]),
						   GNOME_MONTH_ITEM_DAY_LABEL + day_index);
		gnome_canvas_item_set (item,
				       "fill_color", color,
				       "font", NORMAL_DAY_FONT,
				       NULL);

		yv->old_marked_day = -1;
	}

	/* Mark the new day */

	t = time (NULL);
	tm = *localtime (&t);

	if ((tm.tm_year + 1900) == yv->year) {
		const gchar *color;

		color = yv->priv->color_current;
		if (!pango_color_parse (NULL, color))
			color = "#000";

		month_index = tm.tm_mon;
		day_index = gnome_month_item_day2index (GNOME_MONTH_ITEM (yv->mitems[month_index]), tm.tm_mday);
		g_assert (day_index != -1);

		item = gnome_month_item_num2child (GNOME_MONTH_ITEM (yv->mitems[month_index]),
						   GNOME_MONTH_ITEM_DAY_LABEL + day_index);
		gnome_canvas_item_set (item,
				       "fill_color", color,
				       "font",       CURRENT_DAY_FONT,
				       NULL);

		yv->old_marked_day = month_index * 42 + day_index;
	}
}

void
year_view_set (YearView *yv, time_t year)
{
	struct tm tm;
	char buf[100];
	int i;

	g_return_if_fail (yv != NULL);
	g_return_if_fail (IS_YEAR_VIEW (yv));

	tm = *localtime (&year);
	yv->year = tm.tm_year + 1900;

	/* Heading */

	sprintf (buf, "%d", yv->year);
	gnome_canvas_item_set (yv->heading,
			       "text", buf,
			       NULL);

	/* Months */

	for (i = 0; i < 12; i++)
		gnome_canvas_item_set (yv->mitems[i],
				       "year", yv->year,
				       "month", i,
				       NULL);

	/* Unmark and re-mark all the months */

	for (i = 0; i < 12; i++) {
		unmark_month_item (GNOME_MONTH_ITEM (yv->mitems[i]));
		mark_month_item (GNOME_MONTH_ITEM (yv->mitems[i]), yv->calendar->cal);
	}

	mark_current_day (yv);
}
