/*
 * Copyright (C) 2002  Sebastian Rittau <srittau@jroger.in-berlin.de>
 * $Id: egg-datetime.c,v 1.1 2002/12/16 13:20:53 srittau Exp $
 *
 * Based on GnomeDateEdit by Miguel de Icaza.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with the Gnome Library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <libintl.h>

#include <stdio.h>
#include <string.h>
#include <time.h>

#include <gdk/gdkkeysyms.h>
#include <gtk/gtk.h>

#include "egg-datetime.h"


#ifndef _
#define _(x) (x)
#endif


/* TODO:
 *  o "now" button in calendar popup?
 *  o timezone support?
 *  o Implement time list as a popup a la Evo?
 */

/* FIXME:
 *  o limit GtkCalendar to clamp times
 */


/* from libgnomeui */
static void
_add_atk_name_desc (GtkWidget *widget, gchar *name, gchar *desc)
{
	AtkObject *aobj;

	g_return_if_fail (GTK_IS_WIDGET (widget));

	aobj = gtk_widget_get_accessible (widget);

	if (name)
		atk_object_set_name (aobj, name);
	if (desc)
		atk_object_set_description (aobj, desc);
}

static void
_add_atk_relation (GtkWidget *widget1, GtkWidget *widget2,
		   AtkRelationType w1_to_w2, AtkRelationType w2_to_w1)
{
	AtkObject      *atk_widget1;
	AtkObject      *atk_widget2;
	AtkRelationSet *relation_set;
	AtkRelation    *relation;
	AtkObject      *targets[1];

	atk_widget1 = gtk_widget_get_accessible (widget1);
	atk_widget2 = gtk_widget_get_accessible (widget2);

	/* Create the widget1 -> widget2 relation */
	relation_set = atk_object_ref_relation_set (atk_widget1);
	targets[0] = atk_widget2;
	relation = atk_relation_new (targets, 1, w1_to_w2);
	atk_relation_set_add (relation_set, relation);
	g_object_unref (relation);

	/* Create the widget2 -> widget1 relation */
	relation_set = atk_object_ref_relation_set (atk_widget2);
	targets[0] = atk_widget1;
	relation = atk_relation_new (targets, 1, w2_to_w1);
	atk_relation_set_add (relation_set, relation);
	g_object_unref (relation);
}

/*
 * Class and Object Handling
 */

struct _EggDateTimePrivate
{
	/* Children */

	GtkWidget *date_box;
	GtkWidget *date_entry;
	GtkWidget *date_button;
	GtkWidget *time_box;
	GtkWidget *time_entry;
	GtkWidget *time_om;
	GtkWidget *cal_popup;
	GtkWidget *calendar;

	/* Flags */

	EggDateTimeDisplayMode display_mode;
	gboolean normalize;
	gboolean week_start_monday;

	/* Current Date/Time */

	guint16 year;
	guint8  month;
	guint8  day;
	guint8  hour;
	guint8  minute;
	guint8  second;

	/* Clamp Values */

	guint16 clamp_minyear,   clamp_maxyear;
	guint8  clamp_minmonth,  clamp_maxmonth;
	guint8  clamp_minday,    clamp_maxday;
	guint8  clamp_minhour,   clamp_maxhour;
	guint8  clamp_minminute, clamp_maxminute;
	guint8  clamp_minsecond, clamp_maxsecond;

	/* Signals */

	gulong time_om_signal;
};

enum {
	SIGNAL_DATE_CHANGED,
	SIGNAL_TIME_CHANGED,
	SIGNAL_LAST
};

static gint egg_datetime_signals [SIGNAL_LAST] = { 0 };


static void	 egg_datetime_class_init	(EggDateTimeClass	*klass);
static void	 egg_datetime_init		(EggDateTime		*edt);

static void	 egg_datetime_destroy		(GtkObject *object);
static void	 egg_datetime_finalize		(GObject		*object);

static gchar	*get_time_string		(guint8 hour, guint8 minute, guint8 second);

static gboolean	 date_focus_out			(EggDateTime *edt,	GtkEntry *entry);
static gboolean	 time_focus_out			(EggDateTime *edt,	GtkEntry *entry);
static void	 date_arrow_toggled		(EggDateTime *edt,	GtkButton *button);
static void	 cal_popup_changed		(EggDateTime *edt,	GtkCalendar *calendar);
static void	 cal_popup_double_click		(EggDateTime *edt,	GtkCalendar *calendar);
static gboolean	 cal_popup_key_pressed		(EggDateTime *edt,	GdkEventKey *event, GtkWidget *widget);
static gboolean	 cal_popup_button_pressed	(EggDateTime *edt,	GdkEventButton *event, GtkWidget *widget);
static gboolean	 cal_popup_closed		(EggDateTime *edt,	GtkWindow *popup);
static void	 time_popup_changed		(EggDateTime *edt,	GtkOptionMenu *om);

static void	 apply_display_mode		(EggDateTime *edt);
static void	 clamp_time_labels		(EggDateTime *edt);
static gboolean	 parse_date			(EggDateTime *edt);
static gboolean	 parse_time			(EggDateTime *edt);
static void	 normalize_date			(EggDateTime *edt);
static void	 normalize_time			(EggDateTime *edt);
static void	 parse_and_normalize_date	(EggDateTime *edt);
static void	 parse_and_normalize_time	(EggDateTime *edt);
static void	 update_date_label		(EggDateTime *edt);
static void	 update_time_label		(EggDateTime *edt);
static void	 cal_popup_hide			(EggDateTime *edt);


static GtkHBoxClass *parent_class = NULL;


GtkType
egg_datetime_get_type (void)
{
	static GtkType datetime_type = 0;

	if (!datetime_type) {
		static const GTypeInfo datetime_info = {
			sizeof (EggDateTimeClass),
			NULL,		/* base_init */
			NULL,		/* base_finalize */
			(GClassInitFunc) egg_datetime_class_init,
			NULL,		/* class_finalize */
			NULL,		/* class_data */
			sizeof (EggDateTime),
			0,		/* n_preallocs */
			(GInstanceInitFunc) egg_datetime_init
		};

		datetime_type = g_type_register_static (GTK_TYPE_HBOX, "EggDateTime", &datetime_info, 0);
	}

	return datetime_type;
}

static void
egg_datetime_class_init (EggDateTimeClass *klass)
{
	GObjectClass *o_class;
	GtkObjectClass *go_class;

	parent_class = g_type_class_peek_parent (klass);

	o_class  = (GObjectClass *)   klass;
	go_class = (GtkObjectClass *) klass;

	o_class->finalize = egg_datetime_finalize;
	go_class->destroy = egg_datetime_destroy;

	/* Signals */

	egg_datetime_signals [SIGNAL_DATE_CHANGED] =
		g_signal_new ("date-changed",
			      G_TYPE_FROM_CLASS (klass),
			      G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
			      G_STRUCT_OFFSET (EggDateTimeClass, date_changed),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__VOID,
			      G_TYPE_NONE, 0);

	egg_datetime_signals [SIGNAL_TIME_CHANGED] =
		g_signal_new ("time-changed",
			      G_TYPE_FROM_CLASS (klass),
			      G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
			      G_STRUCT_OFFSET (EggDateTimeClass, time_changed),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__VOID,
			      G_TYPE_NONE, 0);
}

static void
egg_datetime_init (EggDateTime *edt)
{
	GtkWidget *arrow, *menu;
	GtkCalendarDisplayOptions cal_options;
	gint i;

	edt->priv = g_new0 (EggDateTimePrivate, 1);

	edt->priv->display_mode = EGG_DATETIME_DISPLAY_DATE;
	edt->priv->normalize = TRUE;

	/* Translators: Change this if you want to start weeks at mondays. */
	edt->priv->week_start_monday = strcmp (_("week-starts-monday: no"), "week-starts-monday: no") ? TRUE : FALSE;

	edt->priv->clamp_minyear   = 1;
	edt->priv->clamp_minmonth  = 1;
	edt->priv->clamp_minday    = 1;
	edt->priv->clamp_maxyear   = 9999;
	edt->priv->clamp_maxmonth  = 12;
	edt->priv->clamp_maxday    = 31;
	edt->priv->clamp_minhour   = 0;
	edt->priv->clamp_minminute = 0;
	edt->priv->clamp_minsecond = 0;
	edt->priv->clamp_maxhour   = 23;
	edt->priv->clamp_maxminute = 59;
	edt->priv->clamp_maxsecond = 59;

	/* Initialize Widgets */

	gtk_box_set_spacing (GTK_BOX (edt), 4);

	/* Date Widgets */

	edt->priv->date_box = gtk_hbox_new (FALSE, 0);
	gtk_box_pack_start (GTK_BOX (edt), edt->priv->date_box, TRUE, TRUE, 0);

	edt->priv->date_entry = gtk_entry_new ();
	gtk_entry_set_width_chars (GTK_ENTRY (edt->priv->date_entry), 12);
	_add_atk_name_desc (edt->priv->date_entry, _("Date"), _("Enter the date directly"));
	g_signal_connect_swapped (G_OBJECT (edt->priv->date_entry), "focus-out-event",
				  G_CALLBACK (date_focus_out), edt);
	gtk_widget_show (edt->priv->date_entry);
	gtk_box_pack_start (GTK_BOX (edt->priv->date_box), edt->priv->date_entry, TRUE, TRUE, 0);

	edt->priv->date_button = gtk_toggle_button_new ();
	_add_atk_name_desc (edt->priv->date_button, _("Select Date"), _("Select the date from a calendar"));
	gtk_box_pack_start (GTK_BOX (edt->priv->date_box), edt->priv->date_button, FALSE, FALSE, 0);
	arrow = gtk_arrow_new (GTK_ARROW_DOWN, GTK_SHADOW_OUT);
	gtk_container_add (GTK_CONTAINER (edt->priv->date_button), arrow);
	gtk_widget_show (arrow);
	g_signal_connect_swapped (G_OBJECT (edt->priv->date_button), "toggled",
				  G_CALLBACK (date_arrow_toggled), edt);

	_add_atk_relation (edt->priv->date_button, edt->priv->date_entry,
			   ATK_RELATION_CONTROLLER_FOR, ATK_RELATION_CONTROLLED_BY);

	/* Time Widgets */

	edt->priv->time_box = gtk_hbox_new (FALSE, 0);
	gtk_box_pack_start (GTK_BOX (edt), edt->priv->time_box, TRUE, TRUE, 0);

	edt->priv->time_entry = gtk_entry_new ();
	gtk_entry_set_width_chars (GTK_ENTRY (edt->priv->time_entry), 10);
	_add_atk_name_desc (edt->priv->time_entry, _("Time"), _("Enter the time directly"));
	g_signal_connect_swapped (G_OBJECT (edt->priv->time_entry), "focus-out-event",
				  G_CALLBACK (time_focus_out), edt);
	gtk_box_pack_start (GTK_BOX (edt->priv->time_box), edt->priv->time_entry, TRUE, TRUE, 0);
	gtk_widget_show (edt->priv->time_entry);

	edt->priv->time_om = gtk_option_menu_new ();
	gtk_box_pack_start (GTK_BOX (edt->priv->time_box), edt->priv->time_om, FALSE, TRUE, 0);
	edt->priv->time_om_signal = 
		g_signal_connect_swapped (G_OBJECT (edt->priv->time_om), "changed",
					  G_CALLBACK (time_popup_changed), edt);

	menu = gtk_menu_new ();
	gtk_widget_show (menu);
	for (i = 0; i <= 24; i++) {
		GtkWidget *item;
		gchar *s;

		s = get_time_string (i, 0, 0xff);
		item = gtk_menu_item_new_with_label (s);
		g_free (s);
		gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
	}
	gtk_option_menu_set_menu (GTK_OPTION_MENU (edt->priv->time_om), menu);
	clamp_time_labels (edt);

	/* Calendar Popup */

	edt->priv->cal_popup = gtk_window_new (GTK_WINDOW_POPUP);
	gtk_widget_set_events (edt->priv->cal_popup,
			       gtk_widget_get_events (edt->priv->cal_popup) | GDK_KEY_PRESS_MASK);
	gtk_window_set_resizable (GTK_WINDOW (edt->priv->cal_popup), FALSE);
	g_signal_connect_swapped (G_OBJECT (edt->priv->cal_popup), "delete-event",
				  G_CALLBACK (cal_popup_closed), edt);
	g_signal_connect_swapped (G_OBJECT (edt->priv->cal_popup), "key-press-event",
				  G_CALLBACK (cal_popup_key_pressed), edt);
	g_signal_connect_swapped (G_OBJECT (edt->priv->cal_popup), "button-press-event",
				  G_CALLBACK (cal_popup_button_pressed), edt);

	edt->priv->calendar = gtk_calendar_new ();
	cal_options = GTK_CALENDAR_SHOW_DAY_NAMES | GTK_CALENDAR_SHOW_HEADING;
	if (edt->priv->week_start_monday)
		cal_options |= GTK_CALENDAR_WEEK_START_MONDAY;
	gtk_calendar_display_options (GTK_CALENDAR (edt->priv->calendar), cal_options);
        gtk_container_add (GTK_CONTAINER (edt->priv->cal_popup), edt->priv->calendar);
	g_signal_connect_swapped (G_OBJECT (edt->priv->calendar), "day-selected",
				  G_CALLBACK (cal_popup_changed), edt);
	g_signal_connect_swapped (G_OBJECT (edt->priv->calendar), "day-selected-double-click",
				  G_CALLBACK (cal_popup_double_click), edt);
        gtk_widget_show (edt->priv->calendar);

	apply_display_mode (edt);
}

static void
egg_datetime_destroy (GtkObject *object)
{
	EggDateTime *edt = EGG_DATETIME (object);

	if (edt->priv->cal_popup) {
		gtk_widget_destroy (edt->priv->cal_popup);
		edt->priv->cal_popup = NULL;
	}

	if (GTK_OBJECT_CLASS (parent_class)->destroy)
		(* GTK_OBJECT_CLASS (parent_class)->destroy) (object);
}

static void
egg_datetime_finalize (GObject *object)
{
	EggDateTime *edt = EGG_DATETIME (object);

	g_free (edt->priv);

	if (G_OBJECT_CLASS (parent_class)->finalize)
		(* G_OBJECT_CLASS (parent_class)->finalize) (object);
}

/*
 * Utility Functions
 */

/* Determine the number of bits, time_t uses. */
static guint
time_t_bits (void) {
	guint i;
	time_t t;

	for (t = 1, i = 0; t != 0; t = t << 1, i++)
		;

	return i;
}

static gchar *
get_time_string (guint8 hour, guint8 minute, guint8 second)
{
	gchar *s;

	/* Translators: set this to anything else if you want to use a
	 * 24 hour clock.
	 */
	if (!strcmp (_("24hr: no"), "24hr: no")) {
		const gchar *ampm_s;

		if (hour < 12)
			ampm_s = _("AM");
		else
			ampm_s = _("PM");

		hour %= 12;
		if (hour == 0)
			hour = 12;

		if (second <= 59)
			/* Translators: This is hh:mm:ss AM/PM. */
			s = g_strdup_printf (_("%02d:%02d:%02d %s"), hour, minute, second, ampm_s);
		else
			/* Translators: This is hh:mm AM/PM. */
			s = g_strdup_printf (_("%02d:%02d %s"), hour, minute, ampm_s);
	} else {
		if (second <= 59)
			/* Translators: This is hh:mm:ss. */
			s = g_strdup_printf (_("%02d:%02d:%02d"), hour, minute, second);
		else
			/* Translators: This is hh:mm. */
			s = g_strdup_printf (_("%02d:%02d"), hour, minute);
	}

	return s;
}

/*
 * Callbacks
 */

static gboolean
date_focus_out (EggDateTime *edt, GtkEntry *entry)
{
	gboolean ret;

	ret = parse_date (edt);
	normalize_date (edt);
	if (ret || edt->priv->normalize)
		update_date_label (edt);

	g_signal_emit (G_OBJECT (edt), egg_datetime_signals[SIGNAL_DATE_CHANGED], 0);

	return TRUE;
}

static gboolean
time_focus_out (EggDateTime *edt, GtkEntry *entry)
{
	gboolean ret;

	ret = parse_time (edt);
	normalize_time (edt);
	if (ret || edt->priv->normalize)
		update_time_label (edt);

	g_signal_emit (G_OBJECT (edt), egg_datetime_signals[SIGNAL_TIME_CHANGED], 0);

	return TRUE;
}

static void
date_arrow_toggled (EggDateTime *edt, GtkButton *button)
{
	GdkCursor *cursor;
	GdkWindow *window;
	GtkRequisition requisition;
	gint x, y, width, height;
	guint16 year;
	guint8 month, day;

	if (!gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (edt->priv->date_button)))
		return;

	/* Set Date */

	if (parse_date (edt)) {
		year  = edt->priv->year;
		month = edt->priv->month;
		day   = edt->priv->day;
	} else {
		struct tm tm;
		time_t t;

		t = time (NULL);
		localtime_r (&t, &tm);

		year  = tm.tm_year + 1900;
		month = tm.tm_mon + 1;
		day   = tm.tm_mday;
	}
	gtk_calendar_select_month (GTK_CALENDAR (edt->priv->calendar), month - 1, year);
	gtk_calendar_select_day (GTK_CALENDAR (edt->priv->calendar), day);

	/* Position Popup */

	window = edt->priv->date_button->window;
	gtk_widget_size_request (edt->priv->cal_popup, &requisition);
	gdk_window_get_origin (window, &x, &y);

	x += edt->priv->date_button->allocation.x;
	y += edt->priv->date_button->allocation.y;
	width  = edt->priv->date_button->allocation.width;
	height = edt->priv->date_button->allocation.height;

	x += width - requisition.width;
	y += height;

	if (x < 0)
		x = 0;
	if (y < 0)
		y = 0;

	gtk_window_move (GTK_WINDOW (edt->priv->cal_popup), x, y);

	/* Show Popup */

	gtk_widget_show (edt->priv->cal_popup);
	gtk_widget_grab_focus (edt->priv->cal_popup);
	gtk_grab_add (edt->priv->cal_popup);

	cursor = gdk_cursor_new (GDK_ARROW);

	gdk_pointer_grab (edt->priv->cal_popup->window, TRUE,
			  (GDK_BUTTON_PRESS_MASK
			   | GDK_BUTTON_RELEASE_MASK
			   | GDK_POINTER_MOTION_MASK),
			  NULL, cursor, GDK_CURRENT_TIME);

	gdk_cursor_destroy (cursor);
}

static void
cal_popup_changed (EggDateTime *edt, GtkCalendar *calendar)
{
	guint year, month, day;

	gtk_calendar_get_date (GTK_CALENDAR (edt->priv->calendar), &year, &month, &day);

	edt->priv->year  = year;
	edt->priv->month = month + 1;
	edt->priv->day   = day;

	update_date_label (edt);

	g_signal_emit (G_OBJECT (edt), egg_datetime_signals[SIGNAL_DATE_CHANGED], 0);
}

static void
cal_popup_double_click (EggDateTime *edt, GtkCalendar *calendar)
{
	cal_popup_hide (edt);
}

static gboolean
cal_popup_key_pressed (EggDateTime *edt, GdkEventKey *event, GtkWidget *widget)
{
	if (event->keyval != GDK_Escape)
		return FALSE;

	g_signal_stop_emission_by_name (G_OBJECT (widget), "key_press_event");

	cal_popup_hide (edt);

	return TRUE;
}

/* Stolen from gtkcombo.c */
static gboolean
cal_popup_button_pressed (EggDateTime *edt, GdkEventButton *event, GtkWidget *widget)
{
	GtkWidget *child;

	child = gtk_get_event_widget ((GdkEvent *) event);

	/* We don't ask for button press events on the grab widget, so
	 * if an event is reported directly to the grab widget, it must
	 * be on a window outside the application (and thus we remove
	 * the popup window). Otherwise, we check if the widget is a child
	 * of the grab widget, and only remove the popup window if it
	 * is not.
	 */
	if (child != widget) {
		while (child) {
			if (child == widget)
				return FALSE;
			child = child->parent;
		}
	}

	cal_popup_hide (edt);

	return TRUE;
}

static gboolean
cal_popup_closed (EggDateTime *edt, GtkWindow *popup)
{
	cal_popup_hide (edt);

	return TRUE;
}

static void
time_popup_changed (EggDateTime *edt, GtkOptionMenu *om)
{
	gint i;

	i = gtk_option_menu_get_history (om);

	edt->priv->hour   = i;
	edt->priv->minute = 0;
	edt->priv->second = 0;

	update_time_label (edt);

	g_signal_emit (G_OBJECT (edt), egg_datetime_signals[SIGNAL_TIME_CHANGED], 0);
}

/*
 * Private Methods
 */

static void
apply_display_mode (EggDateTime *edt)
{
	if (edt->priv->display_mode & EGG_DATETIME_DISPLAY_DATE)
		gtk_widget_show (edt->priv->date_box);
	else
		gtk_widget_hide (edt->priv->date_box);
	if (edt->priv->display_mode & EGG_DATETIME_DISPLAY_MONTH)
		gtk_widget_show (edt->priv->date_button);
	else
		gtk_widget_hide (edt->priv->date_button);

	if (edt->priv->display_mode & EGG_DATETIME_DISPLAY_TIME)
		gtk_widget_show (edt->priv->time_box);
	else
		gtk_widget_hide (edt->priv->time_box);
	if (edt->priv->display_mode & EGG_DATETIME_DISPLAY_HOUR)
		gtk_widget_show (edt->priv->time_om);
	else
		gtk_widget_hide (edt->priv->time_om);
}

static void
clamp_time_labels (EggDateTime *edt)
{
	guint min, max;

	min = edt->priv->clamp_minhour;
	if (edt->priv->clamp_minminute > 0)
		min++;
	max = edt->priv->clamp_maxhour;

	if (min > 23 || max - min < 1)
		gtk_widget_hide (edt->priv->time_om);
	else {
		GtkWidget *menu;
		GList *list;
		gint i;

		menu = gtk_option_menu_get_menu (GTK_OPTION_MENU (edt->priv->time_om));

		/* Show all items in range of [min;max], hide all others */

		list = GTK_MENU_SHELL (menu)->children;
		for (i = 0; i <= 23; i++, list = g_list_next (list)) {
			GtkWidget *item;

			g_assert (list != NULL);
			g_assert (list->data != NULL && GTK_IS_WIDGET (list->data));
			item = GTK_WIDGET (list->data);

			if (i >= min && i <= max)
				gtk_widget_show (item);
			else
				gtk_widget_hide (item);
		}

		g_signal_handler_block (G_OBJECT (edt->priv->time_om), edt->priv->time_om_signal);
		i = gtk_option_menu_get_history (GTK_OPTION_MENU (edt->priv->time_om));
		if (i < min)
			gtk_option_menu_set_history (GTK_OPTION_MENU (edt->priv->time_om), min);
		if (i > max)
			gtk_option_menu_set_history (GTK_OPTION_MENU (edt->priv->time_om), max);
		g_signal_handler_unblock (G_OBJECT (edt->priv->time_om), edt->priv->time_om_signal);

		gtk_widget_show (edt->priv->time_om);
	}
}

static void
update_date_label (EggDateTime *edt)
{
	gchar *s;

	/* TODO: should handle other display modes as well... */

	/* Translators: This is yyyy-mm-dd. */
	s = g_strdup_printf (_("%04d-%02d-%02d"), edt->priv->year, edt->priv->month, edt->priv->day);
	gtk_entry_set_text (GTK_ENTRY (edt->priv->date_entry), s);
	g_free (s);
}

static void
update_time_label (EggDateTime *edt)
{
	gchar *s;

	/* TODO: should handle other display modes as well... */

	if ((edt->priv->display_mode & EGG_DATETIME_DISPLAY_SECOND) ||
	    (edt->priv->display_mode & EGG_DATETIME_DISPLAY_SECOND_OPT))
		s = get_time_string (edt->priv->hour, edt->priv->minute, edt->priv->second);
	else
		s = get_time_string (edt->priv->hour, edt->priv->minute, 0xff);

	gtk_entry_set_text (GTK_ENTRY (edt->priv->time_entry), s);

	g_free (s);
}

static gboolean
parse_date (EggDateTime *edt)
{
	GDate date;

	g_date_set_parse (&date, gtk_entry_get_text (GTK_ENTRY (edt->priv->date_entry)));
	if (!g_date_valid (&date))
		return FALSE;

	edt->priv->year  = date.year;
	edt->priv->month = date.month;
	edt->priv->day   = date.day;

	return TRUE;
}

static gboolean
parse_time (EggDateTime *edt)
{
	const gchar *s;
	gchar *scp;
	unsigned int hour, minute, second = 60;
	size_t len1, len2, len3;

	/* Retrieve and Parse String */

	s = gtk_entry_get_text (GTK_ENTRY (edt->priv->time_entry));

	/* Translators: This is hh:mm:ss AM/PM. */
	if (sscanf (s, _("%u:%u:%u"), &hour, &minute, &second) < 2) {
		if (sscanf (s, "%u:%u:%u", &hour, &minute, &second) < 2)
			return FALSE;
	}

	if (hour > 23 || minute > 59 || second > 60)
		return FALSE;

	/* AM/PM Handling */

	scp = g_strdup (s);
	scp = g_strchomp (scp);

	len1 = strlen (_("AM"));
	len2 = strlen (_("PM"));
	len3 = strlen (scp);

	if (len1 < len3 && !strcasecmp (scp + len3 - len1, _("AM"))) {
		if (hour == 12)
			hour = 0;
	}
	if (len2 < len3 && !strcasecmp (scp + len3 - len2, _("PM"))) {
		if (hour == 12)
			hour = 0;
		hour += 12;
	}

	/* Store Time */

	edt->priv->hour   = hour;
	edt->priv->minute = minute;
	if (second <= 59)
		edt->priv->second = second;

	/* Cleanup */

	g_free (scp);

	return TRUE;
}

/* Normalize the current date. */
static void
normalize_date (EggDateTime *edt)
{
	GDate date, min_date, max_date;

	if (!edt->priv->normalize)
		return;

	g_date_clear (&date, 1);
	g_date_set_dmy (&date, edt->priv->day, edt->priv->month, edt->priv->year);
	g_date_clear (&min_date, 1);
	g_date_set_dmy (&min_date, edt->priv->clamp_minday, edt->priv->clamp_minmonth, edt->priv->clamp_minyear);
	g_date_clear (&max_date, 1);
	g_date_set_dmy (&max_date, edt->priv->clamp_maxday, edt->priv->clamp_maxmonth, edt->priv->clamp_maxyear);

	g_date_clamp (&date, &min_date, &max_date);

	edt->priv->year  = date.year;
	edt->priv->month = date.month;
	edt->priv->day   = date.day;
}

/* Normalize the current time. */
static void
normalize_time (EggDateTime *edt)
{
	if (!edt->priv->normalize)
		return;

	if (edt->priv->hour < edt->priv->clamp_minhour) {
		edt->priv->hour   = edt->priv->clamp_minhour;
		edt->priv->minute = edt->priv->clamp_minminute;
		edt->priv->second = edt->priv->clamp_minsecond;
	} else if (edt->priv->hour == edt->priv->clamp_minhour) {
		if (edt->priv->minute < edt->priv->clamp_minminute) {
			edt->priv->minute = edt->priv->clamp_minminute;
			edt->priv->second = edt->priv->clamp_minsecond;
		} else if (edt->priv->minute == edt->priv->clamp_minminute) {
			if (edt->priv->second < edt->priv->clamp_minsecond)
				edt->priv->second = edt->priv->clamp_minsecond;
		}
	}

	if (edt->priv->hour > edt->priv->clamp_maxhour) {
		edt->priv->hour   = edt->priv->clamp_maxhour;
		edt->priv->minute = edt->priv->clamp_maxminute;
		edt->priv->second = edt->priv->clamp_maxsecond;
	} else if (edt->priv->hour == edt->priv->clamp_maxhour) {
		if (edt->priv->minute > edt->priv->clamp_maxminute) {
			edt->priv->minute = edt->priv->clamp_maxminute;
			edt->priv->second = edt->priv->clamp_maxsecond;
		} else if (edt->priv->minute == edt->priv->clamp_maxminute) {
			if (edt->priv->second > edt->priv->clamp_maxsecond)
				edt->priv->second = edt->priv->clamp_maxsecond;
		}
	}
}

static void
parse_and_normalize_date (EggDateTime *edt)
{
	if (!edt->priv->normalize)
		return;

	parse_date (edt);
	normalize_date (edt);
	update_date_label (edt);
}

static void
parse_and_normalize_time (EggDateTime *edt)
{
	if (!edt->priv->normalize)
		return;

	parse_time (edt);
	normalize_time (edt);
	update_time_label (edt);
}

static void
cal_popup_hide (EggDateTime *edt)
{
	gtk_widget_hide (edt->priv->cal_popup);
	gtk_grab_remove (edt->priv->cal_popup);
	gdk_pointer_ungrab (GDK_CURRENT_TIME);
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (edt->priv->date_button), FALSE);
	gtk_widget_grab_focus (edt->priv->date_entry);

	g_signal_emit (G_OBJECT (edt), egg_datetime_signals[SIGNAL_DATE_CHANGED], 0);
}

/*
 * Public Methods
 */

/**
 * egg_datetime_new:
 *
 * Creates a new #EggDateTime widget. By default this widget will show
 * only the date component and is set to the current date and time.
 *
 * Return value: A newly created #EggDateTime widget.
 **/
GtkWidget *
egg_datetime_new (void)
{
	EggDateTime *edt;

	edt = g_object_new (EGG_TYPE_DATETIME, NULL);
	egg_datetime_set_from_time_t (edt, time (NULL));

	return GTK_WIDGET (edt);
}

/**
 * egg_datetime_new_from_time_t:
 *
 * Creates a new #EggDateTime widget and sets it to the date and time
 * given as @t argument. This does also call egg_datetime_set_clamp_time_t().
 * By default this widget will show only the date component.
 *
 * Return value: A newly created #EggDateTime widget.
 **/
GtkWidget *
egg_datetime_new_from_time_t (time_t t)
{
	EggDateTime *edt;

	g_return_val_if_fail (t >= 0, NULL);

	edt = g_object_new (EGG_TYPE_DATETIME, NULL);
	egg_datetime_set_from_time_t (edt, t);
	egg_datetime_set_clamp_time_t (edt);

	return GTK_WIDGET (edt);
}

/**
 * egg_datetime_new_from_struct_tm:
 *
 * Creates a new #EggDateTime widget and sets it to the date and time
 * given as @tm argument. By default this widget will show only the date
 * component.
 *
 * Return value: A newly created #EggDateTime widget.
 **/
GtkWidget *
egg_datetime_new_from_struct_tm (struct tm *tm)
{
	EggDateTime *edt;

	g_return_val_if_fail (tm != NULL, NULL);

	edt = g_object_new (EGG_TYPE_DATETIME, NULL);
	egg_datetime_set_from_struct_tm (edt, tm);

	return GTK_WIDGET (edt);
}

/**
 * egg_datetime_new_from_gdate:
 *
 * Creates a new #EggDateTime widget and sets it to the date and time
 * given as @date argument. By default this widget will show only the
 * date component.
 *
 * Return value: A newly created #EggDateTime widget.
 **/
GtkWidget *
egg_datetime_new_from_gdate (GDate *date)
{
	EggDateTime *edt;

	g_return_val_if_fail (date != NULL, NULL);

	edt = g_object_new (EGG_TYPE_DATETIME, NULL);
	egg_datetime_set_from_gdate (edt, date);

	return GTK_WIDGET (edt);
}

/**
 * egg_datetime_new_from_datetime:
 *
 * Creates a new #EggDateTime widget and sets it to the date and time
 * given as arguments.
 *
 * Return value: A newly created #EggDateTime widget.
 **/
GtkWidget *
egg_datetime_new_from_datetime (guint16 year, guint8 month, guint8 day, guint8 hour, guint8 minute, guint8 second)
{
	EggDateTime *edt;

	edt = g_object_new (EGG_TYPE_DATETIME, NULL);
	egg_datetime_set_date (edt, year, month, day);
	egg_datetime_set_time (edt, hour, minute, second);

	return GTK_WIDGET (edt);
}

/**
 * egg_datetime_set_from_time_t:
 * @edt: An #EggDateTime.
 * @t: Date and time to set the widget to.
 *
 * Sets the date and time of the widget to @t.
 **/
void
egg_datetime_set_from_time_t (EggDateTime *edt, time_t t)
{
	struct tm tm;

	g_return_if_fail (edt != NULL);
	g_return_if_fail (EGG_IS_DATETIME (edt));

	localtime_r (&t, &tm);

	egg_datetime_set_date (edt, tm.tm_year + 1900, tm.tm_mon + 1, tm.tm_mday);
	egg_datetime_set_time (edt, tm.tm_hour, tm.tm_min, tm.tm_sec);
}

/**
 * egg_datetime_get_as_time_t:
 * @edt: An #EggDateTime.
 * @t: Pointer to a %time_t value.
 *
 * Returns the current time as a %time_t value. If the currently entered
 * value is invalid and normalization is turned off or if the entered date
 * can't be represented as a %time_t value, the value is set to -1 and
 * FALSE is returned.
 *
 * Return value: Success indicator.
 **/
gboolean
egg_datetime_get_as_time_t (EggDateTime *edt, time_t *t)
{
	struct tm tm;
	guint16 year;
	guint8 month, day;
	guint8 hour, minute, second;

	g_return_val_if_fail (edt != NULL, FALSE);
	g_return_val_if_fail (EGG_IS_DATETIME (edt), FALSE);

	if (!t)
		return FALSE;

	if (!egg_datetime_get_date (edt, &year, &month, &day)) {
		*t = -1;
		return FALSE;
	}
	if (!egg_datetime_get_time (edt, &hour, &minute, &second)) {
		*t = -1;
		return FALSE;
	}

	memset (&tm, 0, sizeof (struct tm));
	tm.tm_year = year - 1900;
	tm.tm_mon  = month - 1;
	tm.tm_mday = day;
	tm.tm_hour = hour;
	tm.tm_min  = minute;
	tm.tm_sec  = second;

	*t = mktime (&tm);

	if (*t < 0) {
		*t = -1;
		return FALSE;
	}
	
	return TRUE;
}

/**
 * egg_datetime_set_from_struct_tm:
 * @edt: An #EggDateTime.
 * @tm: Date and time to set the widget to.
 *
 * Sets the date and time of the widget to @tm.
 **/
void
egg_datetime_set_from_struct_tm (EggDateTime *edt, struct tm *tm)
{
	g_return_if_fail (edt != NULL);
	g_return_if_fail (EGG_IS_DATETIME (edt));
	g_return_if_fail (tm != NULL);

	egg_datetime_set_date (edt, tm->tm_year + 1900, tm->tm_mon + 1, tm->tm_mday);
	egg_datetime_set_time (edt, tm->tm_hour, tm->tm_min, tm->tm_sec);
}

/**
 * egg_datetime_get_as_struct_tm:
 * @edt: An #EggDateTime.
 * @tm: Pointer to an allocated struct tm.
 *
 * Fill the supplied struct tm with the widget's current date and time.
 * If the currently entered value is invalid and normalization is turned
 * off or if the entered date can't be represented as a struct tm, the
 * struct is filled with invalid data and FALSE is returned.
 *
 * Return value: Success indicator.
 **/
gboolean
egg_datetime_get_as_struct_tm (EggDateTime *edt, struct tm *tm)
{
	guint16 year;
	guint8 month, day;
	guint8 hour, minute, second;

	g_return_val_if_fail (edt != NULL, FALSE);
	g_return_val_if_fail (EGG_IS_DATETIME (edt), FALSE);

	if (!tm)
		return FALSE;

	memset (tm, 0, sizeof (struct tm));

	if (!egg_datetime_get_date (edt, &year, &month, &day))
		return FALSE;
	if (!egg_datetime_get_time (edt, &hour, &minute, &second))
		return FALSE;

	tm->tm_year = year - 1900;
	tm->tm_mon  = month - 1;
	tm->tm_mday = day;
	tm->tm_hour = hour;
	tm->tm_min  = minute;
	tm->tm_sec  = second;

	mktime (tm);
	
	return TRUE;
}

/**
 * egg_datetime_set_from_gdate:
 * @edt: An #EggDateTime.
 * @date: Date to set the widget to.
 *
 * Sets the date of the widget to @date. The time will remain unchanged.
 **/
void
egg_datetime_set_from_gdate (EggDateTime *edt, GDate *date)
{
	g_return_if_fail (edt != NULL);
	g_return_if_fail (EGG_IS_DATETIME (edt));
	g_return_if_fail (date != NULL);
	g_return_if_fail (g_date_valid (date));

	egg_datetime_set_date (edt, date->year, date->month, date->day);
}

/**
 * egg_datetime_get_as_gdate:
 * @edt: An #EggDateTime.
 * @date: Pointer to an allocated #GDate.
 *
 * Fills the supplied #GDate with the widget's current date. If the
 * currently entered date value is invalid and normalization is turned
 * off or if the entered date can't be represented as a #GDate, the
 * @date is set to an invalid value and FALSE is returned.
 *
 * Return value: Success indicator.
 **/
gboolean
egg_datetime_get_as_gdate (EggDateTime *edt, GDate *date)
{
	guint16 year;
	guint8 month, day;

	g_return_val_if_fail (edt != NULL, FALSE);
	g_return_val_if_fail (EGG_IS_DATETIME (edt), FALSE);

	if (!date)
		return FALSE;

	g_date_clear (date, 1);

	if (!egg_datetime_get_date (edt, &year, &month, &day))
		return FALSE;

	g_date_set_dmy (date, day, month, year);

	return TRUE;
}

/**
 * egg_datetime_set_date:
 * @edt: An #EggDateTime.
 * @year:
 * @month:
 * @day:
 *
 * Sets the date of the widget. The time will remain unchanged.
 **/
void
egg_datetime_set_date (EggDateTime *edt, guint16 year, guint8 month, guint8 day)
{
	g_return_if_fail (edt != NULL);
	g_return_if_fail (EGG_IS_DATETIME (edt));
	g_return_if_fail (year >= 1 && year <= 9999);
	g_return_if_fail (month >= 1 && month <= 12);
	g_return_if_fail (day >= 1 && day <= g_date_get_days_in_month (month, year));

	edt->priv->year  = year;
	edt->priv->month = month;
	edt->priv->day   = day;

	normalize_date (edt);
	update_date_label (edt);

	g_signal_emit (G_OBJECT (edt), egg_datetime_signals[SIGNAL_DATE_CHANGED], 0);
}

/**
 * egg_datetime_get_date:
 * @edt: An #EggDateTime.
 * @year: A pointer to a #guint16 or %NULL.
 * @month: A pointer to #guint8 or %NULL.
 * @day: A pointer to a #guint8 or %NULL.
 *
 * Fills the supplied arguments with the widget's current date. If the
 * currently entered date value is invalid and normalization is turned
 * off, the arguments are set to %EGG_DATETIME_INVALID_DATE and FALSE
 * is returned.
 *
 * Return value: Success indicator.
 **/
gboolean
egg_datetime_get_date (EggDateTime *edt, guint16 *year, guint8 *month, guint8 *day)
{
	g_return_val_if_fail (edt != NULL, FALSE);
	g_return_val_if_fail (EGG_IS_DATETIME (edt), FALSE);

	if (!parse_date (edt)) {
		if (!edt->priv->normalize) {
			if (year)
				*year = 0;
			if (month)
				*month = 0;
			if (day)
				*day = 0;
			return FALSE;
		}
	}

	normalize_date (edt);

	if (year)
		*year  = edt->priv->year;
	if (month)
		*month = edt->priv->month;
	if (day)
		*day   = edt->priv->day;
	
	return TRUE;
}

/**
 * egg_datetime_set_time:
 * @edt: An #EggDateTime.
 * @hour:
 * @minute:
 * @second:
 *
 * Sets the time of the widget. The date will remain unchanged.
 **/
void
egg_datetime_set_time (EggDateTime *edt, guint8 hour, guint8 minute, guint8 second)
{
	g_return_if_fail (edt != NULL);
	g_return_if_fail (EGG_IS_DATETIME (edt));
	g_return_if_fail (hour <= 23);
	g_return_if_fail (minute <= 59);
	g_return_if_fail (second <= 59);

	edt->priv->hour   = hour;
	edt->priv->minute = minute;
	edt->priv->second = second;

	normalize_time (edt);
	update_time_label (edt);

	g_signal_emit (G_OBJECT (edt), egg_datetime_signals[SIGNAL_TIME_CHANGED], 0);
}

/**
 * egg_datetime_get_date:
 * @edt: An #EggDateTime.
 * @hour: A pointer to a #guint8 or %NULL.
 * @minute: A pointer to #guint8 or %NULL.
 * @second: A pointer to a #guint8 or %NULL.
 *
 * Fills the supplied arguments with the widget's current time. If the
 * currently entered time value is invalid and normalization is turned
 * off, the arguments are set to %EGG_DATETIME_INVALID_TIME and FALSE
 * is returned.
 *
 * Return value: Success indicator.
 **/
gboolean
egg_datetime_get_time (EggDateTime *edt, guint8 *hour, guint8 *minute, guint8 *second)
{
	g_return_val_if_fail (edt != NULL, FALSE);
	g_return_val_if_fail (EGG_IS_DATETIME (edt), FALSE);

	if (!parse_time (edt)) {
		if (!edt->priv->normalize) {
			if (hour)
				*hour = 0xff;
			if (minute)
				*minute = 0xff;
			if (second)
				*second = 0xff;
			return FALSE;
		}
	}

	normalize_time (edt);

	if (hour)
		*hour   = edt->priv->hour;
	if (minute)
		*minute = edt->priv->minute;
	if (second)
		*second = edt->priv->second;
	
	return TRUE;
}

/**
 * egg_datetime_set_normalize:
 * @edt: An #EggDateTime.
 * @normalize: A boolean value.
 *
 * If normalize is set to %TRUE, the widget will normalize all invalid
 * values entered in the date and time widgets to the nearest valid
 * value. This guarantees that the get methods will always return valid
 * values.
 *
 * Normalize defaults to %TRUE.
 **/
void
egg_datetime_set_normalize (EggDateTime *edt, gboolean norm)
{
	g_return_if_fail (edt != NULL);
	g_return_if_fail (EGG_IS_DATETIME (edt));

	edt->priv->normalize = norm ? TRUE : FALSE;

	parse_and_normalize_date (edt);
	parse_and_normalize_time (edt);

	g_signal_emit (G_OBJECT (edt), egg_datetime_signals[SIGNAL_DATE_CHANGED], 0);
	g_signal_emit (G_OBJECT (edt), egg_datetime_signals[SIGNAL_TIME_CHANGED], 0);
}

/**
 * egg_datetime_get_normalize:
 * @edt: An #EggDateTime.
 *
 * Return whether values entered into the date and time widgets are
 * normalized to the nearest valid value.
 *
 * Return value: A boolean value.
 **/
gboolean
egg_datetime_get_normalize (EggDateTime *edt)
{
	g_return_val_if_fail (edt != NULL, FALSE);
	g_return_val_if_fail (EGG_IS_DATETIME (edt), FALSE);

	return edt->priv->normalize;
}

/**
 * egg_datetime_set_display_mode:
 * @edt: An #EggDateTime.
 * @mode: New display mode.
 *
 * Sets the widget's new display mode to @mode. The display mode defaults
 * to %EGG_DATETIME_DISPLAY_DATE.
 **/
void
egg_datetime_set_display_mode (EggDateTime *edt, EggDateTimeDisplayMode mode)
{
	g_return_if_fail (edt != NULL);
	g_return_if_fail (EGG_IS_DATETIME (edt));

	edt->priv->display_mode = mode;

	apply_display_mode (edt);
}

/**
 * egg_datetime_get_display_mode:
 * @edt: An #EggDateTime.
 *
 * Returns the current display mode.
 *
 * Return value: The current display mode.
 **/
EggDateTimeDisplayMode
egg_datetime_get_display_mode (EggDateTime *edt)
{
	g_return_val_if_fail (edt != NULL, 0);
	g_return_val_if_fail (EGG_IS_DATETIME (edt), 0);

	return edt->priv->display_mode;
}

/**
 * egg_datetime_set_clamp_date:
 * @edt: An #EggDateTime.
 * @minyear: Minimum year.
 * @minmonth: Minimum month.
 * @minday: Minimum day.
 * @maxyear: Maximum year.
 * @maxmonth: Maximum month.
 * @maxday: Maximum day.
 *
 * Limits the allowed dates to the range given. If normalization is
 * turned on, dates that are outside of this range are snapped to the
 * minimum or maximum date. Otherwise such dates return an invalid value.
 *
 * This defaults to the minimum date 1-1-1 and maximum date 9999-12-31.
 * The maximum year is always limited to 9999.
 **/
void
egg_datetime_set_clamp_date (EggDateTime *edt, guint16 minyear, guint8 minmonth, guint8 minday, guint16 maxyear, guint8 maxmonth, guint8 maxday)
{
	if (maxyear > 9999)
		maxyear = 9999;

	g_return_if_fail (minyear >= 1 && minyear <= 9999 && maxyear >= 1);
	g_return_if_fail (minmonth >= 1 && minmonth <= 12 && maxmonth >= 1 && maxmonth <= 12);
	g_return_if_fail (minday >= 1 && minday <= g_date_get_days_in_month (minmonth, minyear));
	g_return_if_fail (maxday >= 1 && maxday <= g_date_get_days_in_month (maxmonth, maxyear));
	g_return_if_fail (minyear <= maxyear);
	g_return_if_fail (minyear < maxyear || minmonth <= maxmonth);
	g_return_if_fail (minyear < maxyear || minmonth < maxmonth || minday <= maxday);

	edt->priv->clamp_minyear  = minyear;
	edt->priv->clamp_minmonth = minmonth;
	edt->priv->clamp_minday   = minday;
	edt->priv->clamp_maxyear  = maxyear;
	edt->priv->clamp_maxmonth = maxmonth;
	edt->priv->clamp_maxday   = maxday;

	parse_and_normalize_date (edt);

	g_signal_emit (G_OBJECT (edt), egg_datetime_signals[SIGNAL_DATE_CHANGED], 0);
}

/**
 * egg_datetime_set_clamp_time:
 * @edt: An #EggDateTime.
 * @minhour: Minimum hour or 0xff.
 * @minminute: Minimum minute or 0xff.
 * @minsecond: Minimum second or 0xff.
 * @maxhour: Maximum hour or 0xff.
 * @maxminute: Maximum minute or 0xff.
 * @maxsecond: Maximum second or 0xff.
 *
 * Limits the allowed times to the range given. A value of 0xff means
 * an unlimited range. If normalization is turned on, times that are
 * outside of this range are snapped to the minimum or maximum time.
 * Otherwise such times return an invalid value.
 **/
void
egg_datetime_set_clamp_time (EggDateTime *edt, guint8 minhour, guint8 minminute, guint8 minsecond, guint8 maxhour, guint8 maxminute, guint8 maxsecond)
{
	g_return_if_fail (minhour <= 23 && maxhour <= 23);
	g_return_if_fail (minminute <= 59 && maxminute <= 59);
	g_return_if_fail (minsecond <= 59 && maxsecond <= 59);
	g_return_if_fail (minhour <= maxhour);
	g_return_if_fail (minhour < maxhour || minminute <= maxminute);
	g_return_if_fail (minhour < maxhour || minminute < maxminute || minsecond <= maxsecond);

	edt->priv->clamp_minhour   = minhour;
	edt->priv->clamp_minminute = minminute;
	edt->priv->clamp_minsecond = minsecond;
	edt->priv->clamp_maxhour   = maxhour;
	edt->priv->clamp_maxminute = maxminute;
	edt->priv->clamp_maxsecond = maxsecond;

	clamp_time_labels (edt);
	parse_and_normalize_time (edt);

	g_signal_emit (G_OBJECT (edt), egg_datetime_signals[SIGNAL_TIME_CHANGED], 0);
}

/**
 * egg_datetime_set_clamp_time_t:
 * @edt: An #EggDateTime.
 *
 * Clamps the allowed dates of the widget to valid #time_t values.
 * The time clamp settings are not changed.
 **/
void
egg_datetime_set_clamp_time_t (EggDateTime *edt)
{
	time_t t;
	struct tm start_tm, end_tm;
	guint bits;
	guint16 year;
	guint8 month, day;

	g_return_if_fail (edt != NULL);
	g_return_if_fail (EGG_IS_DATETIME (edt));

	t = 0;
	gmtime_r (&t, &start_tm);

	/* evil hack */
	bits = time_t_bits ();
	t = ~0;
	t &= ~(1 << (bits - 1));

	gmtime_r (&t, &end_tm);

	/* Subtract one day from the end date, since not all times of
	 * the last day can be represented.
	 */

	year  = end_tm.tm_year + 1900;
	month = end_tm.tm_mon + 1;
	day   = end_tm.tm_mday;

	if (--day == 0)	{
		if (--month == 0) {
			year--;
			month = 12;
		}

		day = g_date_get_days_in_month (month, year);
	}

	egg_datetime_set_clamp_date (edt, start_tm.tm_year + 1900, start_tm.tm_mon + 1, start_tm.tm_mday, year, month, day);
}

/**
 * egg_datetime_get_clamp_date:
 * @edt: An #EggDateTime.
 * @minyear: #guint16 pointer or %NULL.
 * @minmonth: #guint8 pointer or %NULL.
 * @minday: #guint8 pointer or %NULL.
 * @maxyear: #guint16 pointer or %NULL.
 * @maxmonth: #guint8 pointer or %NULL.
 * @maxday: #guint8 pointer or %NULL.
 *
 * Returns the current clamp date setting.
 **/
void
egg_datetime_get_clamp_date (EggDateTime *edt, guint16 *minyear, guint8 *minmonth, guint8 *minday, guint16 *maxyear, guint8 *maxmonth, guint8 *maxday)
{
	g_return_if_fail (edt != NULL);
	g_return_if_fail (EGG_IS_DATETIME (edt));

	if (minyear)
		*minyear = edt->priv->clamp_minyear;
	if (minmonth)
		*minmonth = edt->priv->clamp_minmonth;
	if (minday)
		*minday = edt->priv->clamp_minday;
	if (maxyear)
		*maxyear = edt->priv->clamp_maxyear;
	if (maxmonth)
		*maxmonth = edt->priv->clamp_maxmonth;
	if (maxday)
		*maxday = edt->priv->clamp_maxday;
}

/**
 * egg_datetime_get_clamp_time:
 * @edt: An #EggDateTime.
 * @minhour: #guint8 pointer or %NULL.
 * @minminute: #guint8 pointer or %NULL.
 * @minsecond: #guint8 pointer or %NULL.
 * @maxhour: #guint8 pointer or %NULL.
 * @maxminute: #guint8 pointer or %NULL.
 * @maxsecond: #guint8 pointer or %NULL.
 *
 * Returns the current clamp time setting.
 **/
void
egg_datetime_get_clamp_time (EggDateTime *edt, guint8 *minhour, guint8 *minminute, guint8 *minsecond, guint8 *maxhour, guint8 *maxminute, guint8 *maxsecond)
{
	g_return_if_fail (edt != NULL);
	g_return_if_fail (EGG_IS_DATETIME (edt));

	if (minhour)
		*minhour = edt->priv->clamp_minhour;
	if (minminute)
		*minminute = edt->priv->clamp_minminute;
	if (minsecond)
		*minsecond = edt->priv->clamp_minsecond;
	if (maxhour)
		*maxhour = edt->priv->clamp_maxhour;
	if (maxminute)
		*maxminute = edt->priv->clamp_maxminute;
	if (maxsecond)
		*maxsecond = edt->priv->clamp_maxsecond;
}
